<?php
namespace Folha\Logic\ESocial;

use Folha\Entities\GpeEsocial;
use HCM\Util\DateUtil;
use Importacao\Logic\Bases\ModeloImportacao\DataBuilder;

class GeradorEventos 
{

    private $dsordem;
    private $limiteInferior;
    private $limiteSuperior;
    private $grupoEventos;
    private $itensEventos;
    private $periodicosEnviados;
    private $versao;
    private $versaoHRM;
    private $tpAmbiente;
    private $nrorg;

    private $dateTime;
    private $caminhoArquivos;
    private $diretorioVersao;

    /**
     * @param $dsordem          Formato de filtragem dos registros a serem gerados (Alfabética, Matrícula, Data de Admissão, Estrutura).
     * @param $limiteInferior   Limite inferior para filtragem dos registros.
     * @param $limiteSuperior   Limite superior para filtragem dos registros.
     * @param $grupoEventos     Código do grupo de eventos que estão sendo gerados (Iniciais, Tabelas, Periódicos, Não Periódicos).
     * @param $itensEventos     Lista com o nome dos eventos a serem gerados. (Ex.: S-1000, S-1060, S-2100)
     * @param $versao           Versão do E-Social a ser gerada. (Ex.: 1.1, 1.2)
     * @param $nrorg            Organização.
     * @param $inicioPeriodo    Data inicial do período considerado na geração de eventos periódicos. Para os outros grupos de eventos a data é desconsiderada.
     * @param $fimPeriodo       Data final do período considerado na geração de eventos periódicos. Para os outros grupos de eventos a data é considerada como data única.
     */
    function __construct($dsordem, $limiteInferior, $limiteSuperior, $grupoEventos, $itensEventos,
                         $versao, $nrorg, $inicioPeriodo, $fimPeriodo)
    {
        $this->dsordem        = $dsordem;
        $this->limiteInferior = $limiteInferior;
        $this->limiteSuperior = $limiteSuperior;
        $this->grupoEventos   = $grupoEventos;
        $this->itensEventos   = $itensEventos;
        $this->versao         = $versao;
        $this->versaoHRM      = "1.0";
        $this->tpAmbiente     = 1; /*Identificação do Ambiente: 1->Produção ; 2->Pré-Produção - dados reais ; 3->Pré-Produção - dados fictícios*/
        $this->nrorg          = $nrorg;
        $this->inicioPeriodo  = $inicioPeriodo;
        $this->fimPeriodo     = $fimPeriodo;

        $this->cdoperador     = \Geral\Util\Ambiente::getCdOperador();

        switch($grupoEventos){
            case 'I':
                $nmGrupoEventos = 'Eventos Iniciais';
                break;
            case 'T':
                $nmGrupoEventos = 'Eventos de Tabela';
                break;
            case 'P':
                $nmGrupoEventos = 'Eventos Periódicos';
                break;
            case 'N':
                $nmGrupoEventos = 'Eventos Não Periódicos';
                break;
        }

        $this->periodicosEnviados = array(
            'S-1200' => 'N',
            'S-1300' => 'N',
            'S-1310' => 'N',
            'S-1320' => 'N',
            'S-1330' => 'N',
            'S-1340' => 'N',
            'S-1350' => 'N',
            'S-1360' => 'N',
            'S-1370' => 'N',
            'S-1380' => 'N',
            'S-1390' => 'N',
        );

        $this->dateTime = new \DateTime();
        $this->diretorioVersao = str_replace(".", "-", $this->versao);
        $this->caminhoArquivos = ESOCIAL_GENERATED_PATH."/{$this->nrorg}/{$this->cdoperador}/".
            "{$this->dateTime->format('Y_m_d_H_i_s')}/{$nmGrupoEventos}_v_$this->diretorioVersao";
    }

    /**
     * O método busca os eventos na tabela GPE_ESOCIAL com base nos nomes existentes na lista $this->itensEventos.
     *
     * Para cada evento são carregados os arquivos json da pasta "/files/esocial/...":
     * - $projects: arquivo contendo a lista dos campos que serão buscados para cada evento e seus respectivos valores na estrutura xml
     * - $estrutura: arquivo contendo, para cada evento, os nomes dos campos que serão validados pelo arquivo helper
     * - $helpers: arquivos contendo as regras de validação e as mensagens para cada campo descrito no arquivo estrutura
     *
     * Busca o nome do repositório utilizado e busca os dados no banco para gerar o(s) arquivo(s) xml do evento.
     *
     * @param bool $utilizaIntervaloDeDatas
     * @throws \Exception
     */
    public function geraEventos($utilizaIntervaloDeDatas = true)
    {
        try {
            $eventos = GpeEsocial::getObjectsByWhere("INSTR(?,NOME) > 0 AND VERSAO = ?", array($this->itensEventos, $this->versao));

            if (!is_array($eventos) or is_null($eventos[0])) {
                throw new \Exception("Nenhum evento foi encontrado.");
            }
            foreach ($eventos as $evento) {
                $projects = json_decode(file_get_contents(ESOCIAL_EVENTOS_TKIS_PATH . "/projects.json"), true);
                $helpers = json_decode(file_get_contents(ESOCIAL_EVENTOS_TKIS_PATH . "/helper.json"), true);

                $data = $this->getDadosDoProjeto($projects, $evento, $utilizaIntervaloDeDatas);
                $dataEvento = $this->getDataset($projects, $data, $helpers, $evento);
                $xml = $this->buildXml($dataEvento, $evento);
            }
        } catch (\Exception $e) {
            throw $e;
        }
    }

    /**
     * Gera o arquivo zip com os arquivos xml gerados no diretório temporário.
     */
    public function geraArquivoZip()
    {
        $caminhoArquivos = $this->getDiretorioArquivosGerados();
        $zip     = new \ZipArchive();
        $zipname = "$caminhoArquivos.zip";
        $zip->open($zipname, \ZIPARCHIVE::CREATE);
        $this->adicionaArquivosAoZip($this->getDiretorioArquivosGerados(), $zip);
        $zip->close();
    }

    /**
     * Adiciona todos os arquivos sob um diretorio a um arquivo .zip.
     */
    private function adicionaArquivosAoZip($raiz, \ZipArchive $zip) 
    {
        foreach (glob("$raiz/*") as $arquivo) {
            if (is_dir($arquivo)) {
                $zip->addEmptyDir(end(explode("/", $arquivo)));
                $this->adicionaArquivosAoZip($arquivo, $zip);
            } else {
                $nomeArquivo = explode("/", $arquivo);
                $nomeArquivo = $nomeArquivo[count($nomeArquivo) - 2] . "/" . $nomeArquivo[count($nomeArquivo) - 1];
                $zip->addFile(realpath($arquivo), $nomeArquivo);
            }
        }
    }

    public function getDiretorioArquivosGerados() 
    {
        return $this->caminhoArquivos;
    }

    public function getCaminhoArquivoZip() 
    {
        return "{$this->caminhoArquivos}.zip";
    }

    private function getDadosDoProjeto($projects, $evento, $utilizaIntervaloDeDatas) 
    {
        try {
            $datasource = $projects[$evento->getNome()]["__datasource"];
            $method = $datasource["method"];
            foreach ($datasource["params"] as $name => $value) {
                $method = str_replace($name, "'$value'", $method);
            }

            $method = str_replace("#dsordem", "'$this->dsordem'", $method);
            $method = str_replace("#limiteInferior", "'$this->limiteInferior'", $method);
            $method = str_replace("#limiteSuperior", "'$this->limiteSuperior'", $method);
            $method = str_replace("#inicioPeriodo", "'".DateUtil::getDataDeString($this->inicioPeriodo)."'", $method);
            $method = str_replace("#fimPeriodo", "'".DateUtil::getDataDeString($this->fimPeriodo)."'", $method);
            //$method = str_replace("#fimPeriodo", "'".DateUtil::truncateString($this->fimPeriodo)."'", $method);
            $method = str_replace("#utilizaIntervaloDeDatas", $utilizaIntervaloDeDatas, $method);
            $method = str_replace("#nrorg", $this->nrorg, $method);
            $method = str_replace("#versaoHRM", $this->versaoHRM, $method);
            //@todo Verificar formato da versao pois segundo o layout deve ser [NN.NN.NN]
            //$versaoEsocial = '0'.$this->versao.'0.00';
            $method = str_replace("#versaoEsocial", $this->versao, $method);
            //$method = str_replace("#versaoEsocial", $versaoEsocial, $method);
            $method = str_replace("#tpAmbiente", $this->tpAmbiente, $method);
            // @todo Verificar idRetificação a ser passado
            $method = str_replace("#indRetificacao", "'N'", $method);
            // @todo Verificar nrRecibo a ser passado
            $method = str_replace("#nrRecibo", '9999', $method);
            // @todo Verificar idApuracao a ser passado
            $method = str_replace("#indApuracao", '1', $method);

            /*Aqui são passados os parâmetros que são utilizados no evento de fechamento dos periódicos, informando quais eventos foram gerados.*/
            if ($evento->getNome() == 'S-1399') {
                $method = str_replace("#evtRemuneracao", $this->periodicosEnviados['S-1200'], $method);
                $method = str_replace("#evtPagamentosDiversos", $this->periodicosEnviados['S-1300'], $method);
                $method = str_replace("#evtServTomadosCMO", $this->periodicosEnviados['S-1310'], $method);
                $method = str_replace("#evtServPrestadosCMO", $this->periodicosEnviados['S-1320'], $method);
                $method = str_replace("#evtServTomadosCOOP", $this->periodicosEnviados['S-1330'], $method);
                $method = str_replace("#evtServPrestadosCOOP", $this->periodicosEnviados['S-1340'], $method);
                $method = str_replace("#evtAquisProdRural", $this->periodicosEnviados['S-1350'], $method);
                $method = str_replace("#evtComercProducao", $this->periodicosEnviados['S-1360'], $method);
                $method = str_replace("#evtRepasseClubeFutebol", $this->periodicosEnviados['S-1370'], $method);
                $method = str_replace("#evtInfoDesoneracao", $this->periodicosEnviados['S-1380'], $method);
                $method = str_replace("#evtReceitaAtivConcomitantes", $this->periodicosEnviados['S-1390'], $method);
            }

            $data = null;
            eval('$data = ' . $method . ';');
            return $data;
        } catch (\Exception $e) {
            throw $e;
        }
    }

    private function getDataset($projects, $data, $helpers, $evento)
    {
        $dataBuilder = new DataBuilder($projects, $data, $helpers);
        $dataBuilder->addAgrupador($evento->getNome());

        return $dataBuilder->build();
    }

    private function buildXml($data, $evento) 
    {
        try {
            $caminhoEventos = $this->caminhoArquivos . "/{$evento->getNome()}";
            if (!is_dir($caminhoEventos)) {
                mkdir($caminhoEventos, 0777, true);
            }

            foreach ($data as $id => $row) {
                $row = $row->getFields();

                /*Busca o campo ID do evento e seta com valor criado conforme o formato na regra de validação REGRA_VALIDA_ID_EVENTO*/
                $keyID = array_search("__ID__", $row, true);
                $idEmpregador = $row['_IDEMPREGADOR'];
                $row[$keyID] = "ID" . $idEmpregador . str_replace("_", "", $this->dateTime->format('Y_m_d_H_i_s')) . str_pad($id + 1, 5, "0", STR_PAD_LEFT);

                $this->verificaTemplate($this->versao, $evento->getNome());
                $templateDir = ESOCIAL_TEMPLATE_ROOT . "/v_$this->diretorioVersao";
                $templateBuilder = new \Geral\Logic\ObrigacoesLegais\Builder\Template($row, $templateDir);
                $xml = $templateBuilder->build("{$evento->getNome()}.xml.twig");
                $nomeArquivo = $evento->getNome() . "_" . str_pad($id + 1, 5, "0", STR_PAD_LEFT) . ".xml";
                file_put_contents("{$this->caminhoArquivos}/{$evento->getNome()}/$nomeArquivo", $xml);

                tk_warning("C:/{$this->caminhoArquivos}/{$evento->getNome()}/$nomeArquivo");

                /*No caso de envio de eventos períodicos são marcados com 'S' os eventos que já foram gerados para informar no fechamento*/
                if ($this->grupoEventos == 'P') {
                    $this->periodicosEnviados[$evento->getNome()] = 'S';
                }

                /*Grava dados na tabela de log de envios do esocial*/
                $this->gerenciaLogEnvios($evento, $row, $xml, $nomeArquivo);
            }
        } catch (\Exception $e) {
            throw $e;
        }
    }

    private function gerenciaLogEnvios($evento, $row, $xml, $nomeArquivo)
    {
        try {
            $logEnvio = new GerenciaEnvios($evento->getNome(), $this->versao, $row['_CHAVE'], $row['_DTMESCOMPETENC'], $this->nrorg);
            $logEnvio->verificaEnvio();
            $arquivoDir = "{$this->caminhoArquivos}/{$evento->getNome()}/$nomeArquivo";
            $logEnvio->setXml($xml, $arquivoDir);

            $logEnvio->persisteEnvio();
        } catch (\Exception $e) {
            throw $e;
        }
    }

    private function verificaTemplate($versao, $evento)
    {
        $arquivo = realpath(ESOCIAL_TEMPLATE_ROOT."/v_$this->diretorioVersao/$evento.xml.twig");
        if (!file_exists($arquivo)) {
            throw new \Folha\Logic\ESocial\Exceptions\VersaoDeArquivoNaoEncontradaException($evento, $versao);
        }
    }

    /**
     * O método busca os eventos na tabela GPE_ESOCIAL com base nos nomes existentes na lista $this->itensEventos.
     *
     * Para cada evento são carregados os arquivos json da pasta "/files/esocial/...":
     * - $projects: arquivo contendo a lista dos campos que serão buscados para cada evento e seus respectivos valores na estrutura xml
     *
     * Busca o nome do repositório utilizado e busca os dados no banco para gerar a visualização de cada evento.
     *
     */
    public function getDataSourceVisualizaEventos()
    {
        try {
            $utilizaIntervaloDeDatas = true;

            $eventos = GpeEsocial::getObjectsByWhere("INSTR(?,NOME) > 0 AND VERSAO = ?", array($this->itensEventos, $this->versao));

            $preVisualizacao = new PreVisualizacao('');

            foreach ($eventos as $key => $evento) {
                $projects = json_decode(file_get_contents(ESOCIAL_EVENTOS_TKIS_PATH . "/projects.json"), true);
                $data = $this->getDadosDoProjeto($projects, $evento, $utilizaIntervaloDeDatas);

                foreach ($data as $item) {
                    $datasource[] = $preVisualizacao->geraVisualizacaoItemEvento($evento->getNome(), $item);
                }
            }

            return $datasource;
        } catch (\Exception $e) {
            throw $e;
        }
    }

    /**
     * O método busca os eventos na tabela GPE_ESOCIAL com base nos nomes existentes na lista $this->itensEventos.
     *
     * Para cada evento são carregados os arquivos json da pasta "/files/esocial/...":
     * - $projects: arquivo contendo a lista dos campos que serão buscados para cada evento e seus respectivos valores na estrutura xml
     * - $estrutura: arquivo contendo, para cada evento, os nomes dos campos que serão validados pelo arquivo helper
     * - $helpers: arquivos contendo as regras de validação e as mensagens para cada campo descrito no arquivo estrutura
     *
     * Executa as regras de validação nos dados de cada evento e retorna um array com os resultados da validação.
     *
     */
    public function getDataSourcePreValidacao()
    {
        $utilizaIntervaloDeDatas = true;

        $eventos = GpeEsocial::getObjectsByWhere("INSTR(?,NOME) > 0 AND VERSAO = '?'", array($this->itensEventos, $this->versao));

        $preValidacao = new PreValidacao();

        foreach ($eventos as $key=>$evento) {
            $projects = json_decode(file_get_contents(ESOCIAL_EVENTOS_TKIS_PATH . "/projects.json"), true);
            $estrutura = json_decode(file_get_contents(ESOCIAL_PRE_VALIDACAO_PATH . "/v_$this->diretorioVersao" . "/estrutura.json"), true);
            $helpers = json_decode(file_get_contents(ESOCIAL_PRE_VALIDACAO_PATH . "/v_$this->diretorioVersao" . "/{$evento->getNome()}_helper.json"), true);
            $data = $this->getDadosDoProjeto($projects, $evento, $utilizaIntervaloDeDatas);

            //$dataValid = $this->getDataset($estrutura, $data, $helpers, $evento);
            $dataValidator = new DataValidator($estrutura, $data, $helpers);
            $dataValidator->addEvento($evento->getNome());
            $dataValid = $dataValidator->build();

            $validacaoEvento = $preValidacao->preValidaEvento($evento->getNome(), $dataValid);

            foreach($validacaoEvento as $erro){
                $datasource[] = $erro;
            }
        }

        return $datasource;
    }
}